﻿using System;
using System.Collections.Generic;
using System.Reflection.Emit;
using RimWorld;
using UnityEngine;
using Verse;
using Verse.Noise;

namespace CharacterEditor
{
  public class EntryInt : EntryObject
  {
    public int valInt;

    public EntryInt(int initialVal = 0, string customControlName = null, bool initialOpen = false)
    {
      valInt = initialVal;
      Initialize(customControlName, initialOpen);
    }

    public override void SetBufferToValue()
    {
      buffer = valInt.ToString();
    }

    public override void ClampValue(float min, float max)
    {
      valInt = (int)Mathf.Clamp(valInt, min, max);
    }

    public override void TryParseBuffer()
    {
      if (float.TryParse(buffer, out float parsedFloat))
      {
        valInt = Mathf.RoundToInt(parsedFloat);
      }
      else if (int.TryParse(buffer, out int parsedInt))
      {
        valInt = parsedInt;
      }
    }

    public override void BackwardButtonAction(float min, float max, bool useCycle)
    {
      valInt--;
      if (valInt < min)
      {
        valInt = useCycle ? (int)max : (int)min;
      }
      SetBufferToValue();
    }

    public override void ForwardButtonAction(float min, float max, bool useCycle)
    {
      valInt++;
      if (valInt > max)
      {
        valInt = useCycle ? (int)min : (int)max;
      }
      SetBufferToValue();
    }

    //public static EntryInt GetEntryForScenPart(ScenPart part, int intialialValue)
    //{
    //  if (!ScenPartDict.TryGetValue(part, out EntryInt entryInt))
    //  {
    //    ////Log.Message("GetEntryForScenPart, initializing " + part);
    //    entryInt = new EntryInt(intialialValue);
    //    ScenPartDict[part] = entryInt;
    //  }
    //  return entryInt;
    //}

    public void UpdateValueAndBuffer(int updatedVal)
    {
      valInt = updatedVal;
      SetBufferToValue();
    }

    public void HorizontalSlider(Rect rect, int min = 0, int max = 100, bool middleAlignment = false, string label = null, string leftAlignedLabel = null, string rightAlignedLabel = null, float roundTo = -1f)
    {
      int valIntClamped = Mathf.Clamp(valInt, min, max);
      int sliderInt = (int)Widgets.HorizontalSlider(rect, valIntClamped, min, max, middleAlignment, label, leftAlignedLabel, rightAlignedLabel, roundTo);
      if (sliderInt != valIntClamped)
      {
        UpdateValueAndBuffer(sliderInt);
      }
    }

    //public void NumericIntBox(Rect rect, int min = 0, int max = 100, bool useArrows = true, bool useCycle = false)
    //{
    //  NumericIntBox(rect.x, rect.y, rect.width, rect.height, min, max, useArrows, useCycle);
    //}

    //public void NumericIntBox(float x, float y, float w = 70f, float h = 26f, int min = 0, int max = 100, bool useArrows = true, bool useCycle = false)
    //{
    //  EntryObjectBox(x, y, w, h, min, max, useArrows, useCycle);
    //}
  }

  public class EntryFloat : EntryObject
  {
    public float valFloat;
    public int roundToDigit;
    private string format;

    public EntryFloat(float initialVal = 0f, string customControlName = null, bool initialOpen = false, int customRounding = 2)
    {
      SetValueRounded(initialVal);
      roundToDigit = customRounding;
      Initialize(customControlName, initialOpen);
      format = "0.";
      for (int i = 0; i < customRounding; i++)
      {
        format += "0";
      }
    }

    public override void SetBufferToValue()
    {
      SetValueRounded(valFloat);
      buffer = valFloat.ToString(format);
    }

    public override void ClampValue(float min, float max)
    {
      SetValueRounded(Mathf.Clamp(valFloat, min, max));
    }

    public override void TryParseBuffer()
    {
      if (float.TryParse(buffer, out float parsedFloat))
      {
        SetValueRounded(parsedFloat);
      }
    }

    public override void BackwardButtonAction(float min, float max, bool useCycle)
    {
      valFloat = valFloat - 1f;
      if (useCycle && valFloat < min)
      {
        valFloat = max;
      }
      SetBufferToValue();
    }

    public override void ForwardButtonAction(float min, float max, bool useCycle)
    {
      valFloat = valFloat + 1f;
      if (useCycle && valFloat > max)
      {
        valFloat = min;
      }
      SetBufferToValue();
    }

    public void UpdateValueAndBuffer(float updatedVal)
    {
      SetValueRounded(updatedVal);
      SetBufferToValue();
    }

    public void SetValueRounded(float rawValue)
    {
      valFloat = (float)Math.Round(rawValue, roundToDigit);
    }

    public void HorizontalSlider(Rect rect, float min = 0, float max = 100)
    {
      float valFloatClamped = Mathf.Clamp(valFloat, min, max);
      float sliderFloat = Widgets.HorizontalSlider(rect, valFloatClamped, min, max, false);
      if (sliderFloat != valFloatClamped)
      {
        UpdateValueAndBuffer(sliderFloat);
      }
    }


    //public EntryFloat(float initialFloat = 0f)
    //{
    //  UpdateValueAndBuffer(initialFloat);
    //}

    //public void UpdateValueAndBuffer(float updatedFloat)
    //{
    //  valFloat = updatedFloat;
    //  SetBufferToValue();
    //}

    //public void SetBufferToValue()
    //{
    //  buffer = valFloat.ToString();
    //}

    //public void ClampValue(float min, float max)
    //{
    //  if (valFloat < min)
    //  {
    //    valFloat = min;
    //  }
    //  else if (max < valFloat)
    //  {
    //    valFloat = max;
    //  }
    //}

    //public void FloatTryParseBuffer()
    //{
    //  if (float.TryParse(buffer, out float parsedFloat))
    //  {
    //    valFloat = parsedFloat;
    //  }
    //}

    //public void TextField(Rect rect, float min = 0f, float max = float.MaxValue)
    //{
    //  if (Mouse.IsOver(rect))
    //  {
    //    ticker = 10;
    //  }
    //  else if (ticker <= 0)
    //  {
    //    if (open)
    //    {
    //      open = false;
    //      ClampValue(min, max);
    //    }
    //    SetBufferToValue();
    //    return;
    //  }
    //  ticker--;
    //  if (!open)
    //  {
    //    open = true;
    //    ClampValue(min, max);
    //    SetBufferToValue();
    //  }
    //  GUI.SetNextControlName(controlName);
    //  buffer = Widgets.TextField(rect, buffer);
    //  if (GUI.GetNameOfFocusedControl() == controlName)
    //  {
    //    ticker = 10;
    //  }
    //  FloatTryParseBuffer();
    //  ClampValue(min, max);
    //}
  }

  public abstract class EntryObject
  {
    public string controlName;
    public static int instanceCounter;
    public string buffer = "";
    public int ticker = 0;
    

    //private const int maxTicker = 5;
    private bool openFlag = false;
    private readonly List<KeyCode> keyCodesToClose = new List<KeyCode>{ KeyCode.Escape , KeyCode.KeypadEnter , KeyCode.Return };

    public bool Open
    {
      get
      {
        return ticker > 0;
      }
      set
      {
        //ticker = value ? maxTicker : 0;
        ticker = value ? 5 : 0;
      }
    }


    public bool CurrentlyFocused => GUI.GetNameOfFocusedControl() == controlName;

    public void Initialize(string customControlName = null, bool initialOpen = false)
    {
      SetBufferToValue();
      Open = initialOpen;
      controlName = customControlName == null ? $"EntryObjectTextField_{instanceCounter++}" : customControlName;
    }

    public void Unfocus()
    {
      if (CurrentlyFocused)
      {
        UI.UnfocusCurrentControl();
      }
    }

    public void Focus()
    {
      GUI.FocusControl(controlName);
    }

    public void TryParseBufferInt(ref int val)
    {
      if (float.TryParse(buffer, out float parsedFloat))
      {
        val = Mathf.RoundToInt(parsedFloat);
      }
    }

    public void TryParseBufferFloat(ref float val)
    {
      if (float.TryParse(buffer, out float parsedFloat))
      {
        val = parsedFloat;
      }
    }

    private bool ControlBoxIntro(Rect rect, float min, float max)
    {
      
      if (CurrentlyFocused && Event.current.type == EventType.KeyDown && keyCodesToClose.Contains(Event.current.keyCode))
      {
        Open = false;
        Unfocus();
        Event.current.Use();
      }
      if (Mouse.IsOver(rect))
      {
        Open = true;
        //ticker = maxTicker;
      }
      else if (OriginalEventUtility.EventType == EventType.MouseDown)
      {
        Open = false;
        Unfocus();
      }
      if (!Open)
      {
        if (openFlag)
        {
          openFlag = false;
          TryParseBuffer();
          ClampValue(min, max);
        }
        SetBufferToValue();
        return false;
      }
      ticker--;
      return true;
    }

    public void NumericTextField(Rect rect, float min, float max, bool useArrows = true, bool useCycle = false)
    {
      NumericTextField(rect.x, rect.y, rect.width, rect.height, min, max, useArrows, useCycle);
    }

    public void NumericTextField(float x, float y, float w, float h, float min, float max, bool useArrows = true, bool useCycle = false)
    {
      Rect rectTotal = new Rect(x, y, w + (useArrows ? 40f : 0f), h);
      if (!ControlBoxIntro(rectTotal, min, max))
      {
        return;
      }
      Rect rectField = new Rect(x + (useArrows ? 20f : 0f), y, w, h);
      if (useArrows)
      {
        Rect rectBack = new Rect(x, y, 25f, h);
        Rect rectForw = new Rect(x + w + 15f, y, 25f, h);
        if (Widgets.ButtonImage(rectBack, ContentFinder<Texture2D>.Get("bbackward", true), true))
        {
          BackwardButtonAction(min, max, useCycle);
        }
        if (Widgets.ButtonImage(rectForw, ContentFinder<Texture2D>.Get("bforward", true), true))
        {
          ForwardButtonAction(min, max, useCycle);
        }
      }
      EntryObjectTextField(rectField, min, max);
    }

    private void EntryObjectTextField(Rect rectTextField, float min, float max)
    {
      if (!openFlag)
      {
        openFlag = true;
        ClampValue(min, max);
        SetBufferToValue();
      }
      //string controlName = "TextFieldInt" + "_x_" + rect.x.ToString("F0") + "_y_" + rect.y.ToString("F0");
      GUI.SetNextControlName(controlName);
      buffer = Widgets.TextField(rectTextField, buffer);
      if (CurrentlyFocused)
      {
        Open = true;
        //ticker = maxTicker;
      }
      TryParseBuffer();
      ClampValue(min, max);
      //if (justOpened)
      //{
      //  Focus();
      //}
    }

    public abstract void SetBufferToValue();

    public abstract void ClampValue(float min, float max);

    public abstract void TryParseBuffer();

    public abstract void BackwardButtonAction(float min, float max, bool useCycle);

    public abstract void ForwardButtonAction(float min, float max, bool useCycle);

    //public object TextFieldObject(object val, Rect rect, object min, object max)
    //{
    //  if (CurrentlyFocused && Event.current.type == EventType.KeyDown && Event.current.keyCode == KeyCode.Escape)
    //  {
    //    Unfocus();
    //    Event.current.Use();
    //  }
    //  if (OriginalEventUtility.EventType == EventType.MouseDown && !rect.Contains(Event.current.mousePosition))
    //  {
    //    Unfocus();
    //  }
    //  if (Mouse.IsOver(rect))
    //  {
    //    ticker = 10;
    //  }
    //  else if (ticker <= 0)
    //  {
    //    if (open)
    //    {
    //      open = false;
    //      ClampValue(ref val, min, max);
    //    }
    //    buffer = val.ToStringSafe();
    //    return val;
    //  }
    //  ticker--;
    //  bool justOpened = false;
    //  if (!open)
    //  {
    //    justOpened = true;
    //    open = true;
    //    ClampValue(ref val, min, max);
    //    buffer = val.ToStringSafe();
    //  }
    //  //string controlName = "TextFieldInt" + "_x_" + rect.x.ToString("F0") + "_y_" + rect.y.ToString("F0");
    //  GUI.SetNextControlName(controlName);
    //  buffer = Widgets.TextField(rect, buffer);
    //  if (CurrentlyFocused)
    //  {
    //    ticker = 10;
    //  }
    //  NumericTryParseBuffer(ref val);
    //  ClampValue(ref val, min, max);
    //  if (justOpened)
    //  {
    //    Focus();
    //  }
    //  return val;

    //}

    //public void ClampValue(ref object val, object min, object max)
    //{
    //  if (val.GetType() == typeof(int))
    //  {
    //    val = Mathf.Clamp((int)val, (int)min, (int)max);
    //    return;
    //  }
    //  if (val is float valFloat)
    //  {
    //    val = Mathf.Clamp(valFloat, (float)min, (float)max);
    //  }
    //}

    //public void NumericTryParseBuffer(ref object val)
    //{
    //  if (float.TryParse(buffer, out float parsedFloat))
    //  {
    //    if (val.GetType() == typeof(int))
    //    {
    //      val = Mathf.RoundToInt(parsedFloat);
    //    }
    //    else if (val is float valFloat)
    //    {
    //      val = valFloat;
    //    }
    //  }
    //}





  }

  //public static class EntryTrackers
  //{
  //  public static Dictionary<string, EntryInt> IntDict = new Dictionary<string, EntryInt>();
  //  public static Dictionary<string, EntryFloat> FloatDict = new Dictionary<string, EntryFloat>();


  //  public static EntryInt GetEntryInt(string entryName, int backupIntitialValue)
  //  {
  //    if (!IntDict.TryGetValue(entryName, out EntryInt entryInt))
  //    {
  //      entryInt = new EntryInt(backupIntitialValue);
  //      IntDict[entryName] = entryInt;
  //    }
  //    return entryInt;
  //  }

  //  public static EntryFloat GetEntryFloat(string entryName, float backupIntitialValue)
  //  {
  //    if (!FloatDict.TryGetValue(entryName, out EntryFloat entryFloat))
  //    {
  //      entryFloat = new EntryFloat(backupIntitialValue);
  //      FloatDict[entryName] = entryFloat;
  //    }
  //    return entryFloat;
  //  } 
  //}

}

