﻿using System;
using System.Collections.Generic;
using System.Linq;
using HarmonyLib;
using RimWorld;
using Verse;
using UnityEngine;

namespace CharacterEditor
{
  public static class GeneTool
  {
    public static char[] separators = new char[] { '|', ':', '?', ';', ',' };

    public static string GetGeneTrackerAsString(this Pawn pawn)
    {
      if (pawn?.genes == null)
      {
        return "";
      }
      Pawn_GeneTracker geneTracker = pawn.genes;
      List<string> geneTrackerStringList = new List<string>();
      geneTrackerStringList.Add("xenogenes" + separators[1] + GetStringFromGeneList(geneTracker.Xenogenes));
      geneTrackerStringList.Add("endogenes" + separators[1] + GetStringFromGeneList(geneTracker.Endogenes));
      geneTrackerStringList.Add("xenotype" + separators[1] + geneTracker.Xenotype?.defName.ToStringSafe());
      geneTrackerStringList.Add("xenotypeName" + separators[1] + geneTracker.xenotypeName.ToStringSafe());
      geneTrackerStringList.Add("hybrid" + separators[1] + geneTracker.hybrid.ToStringSafe());
      geneTrackerStringList.Add("iconDef" + separators[1] + geneTracker.iconDef?.defName.ToStringSafe());
      return String.Join(separators[0].ToString(), geneTrackerStringList);
    }

    public static string GetStringFromGeneList(List<Gene> genes)
    {
      List<string> genesStringList = new List<string>();
      foreach (Gene gene in genes)
      {
        List<string> geneStringList = new List<string>();
        geneStringList.Add("geneDef" + separators[4] + gene?.def?.defName.ToStringSafe());
        geneStringList.Add("loadID" + separators[4] + gene?.loadID.ToStringSafe());
        geneStringList.Add("overriddenByGene" + separators[4] + gene?.overriddenByGene?.loadID.ToStringSafe());
        genesStringList.Add(String.Join(separators[3].ToString(), geneStringList));
      }
      return String.Join(separators[2].ToString(), genesStringList);
    }

    public static List<Tuple<Gene, int?, int?>> GetGeneListFromString(string geneListString, Pawn pawn)
    {
      List<Tuple<Gene, int?, int?>> geneList = new List<Tuple<Gene, int?, int?>>();
      if (geneListString.NullOrEmpty() || geneListString == "null")
      {
        return geneList;
      }
      foreach (string geneString in geneListString.Split(separators[2]))
      {
        Dictionary<string, string> geneDict = new Dictionary<string, string>();
        foreach (string entry in geneString.Split(separators[3]))
        {
          string[] pair = entry.Split(separators[4]);
          if (pair.Count() < 2)
          {
            Log.Warning("GetGeneDataFromString, pair.Count() < 2");
            continue;
          }
          geneDict[pair[0]] = pair[1];
        }
        if (geneDict.TryGetValue("geneDef", out string value) != true)
        {
          continue;
        }
        GeneDef geneDef = DefDatabase<GeneDef>.GetNamedSilentFail(value);
        if (geneDef == null)
        {
          continue;
        }
        int? loadID = ParseIntOrNullFromDict(geneDict, "loadID");
        int? overriddenByGene = ParseIntOrNullFromDict(geneDict, "overriddenByGene");
        geneList.Add(new Tuple<Gene, int?, int?>(GeneMaker.MakeGene(geneDef, pawn), loadID, overriddenByGene));
      }
      return geneList;
    }

    public static int? ParseIntOrNullFromDict(Dictionary<string, string> dict, string key)
    {
      int? result = null;
      if (dict.TryGetValue(key, out string value))
      {
        if (!value.NullOrEmpty() && int.TryParse(value, out int parsedInt))
        {
          result = parsedInt;
        }
      }
      return result;
    }

    public static void SetGeneTracker(this Pawn pawn, string geneTrackerOrMelanin)
    {
      //////Log.Message("SetGeneTracker, start, geneTrackerOrMelanin = " + geneTrackerOrMelanin);
      if (pawn?.genes == null)
      {
        return;
      }
      Pawn_GeneTracker geneTracker = pawn.genes;
      if (float.TryParse(geneTrackerOrMelanin, out float melanin))
      {
        geneTracker.InitializeGenesFromOldSave(melanin);
      }
      else
      {
        string[] array = geneTrackerOrMelanin.Split(separators[0]);
        Dictionary<string, string> geneTrackerDict = new Dictionary<string, string>();
        foreach (string entry in array)
        {
          string[] pair = entry.Split(separators[1]);
          if (pair.Count() < 2)
          {
            Log.Warning("SetGeneTracker, pair.Count() < 2");
            continue;
          }
          geneTrackerDict[pair[0]] = pair[1];
        }
        string value;

        List<Tuple<Gene, int?, int?>> xenogenesData = new List<Tuple<Gene, int?, int?>>();
        if (geneTrackerDict.TryGetValue("xenogenes", out value))
        {
          xenogenesData = GetGeneListFromString(value, pawn);
        }
        List<Tuple<Gene, int?, int?>> endogenesData = new List<Tuple<Gene, int?, int?>>();
        if (geneTrackerDict.TryGetValue("endogenes", out value))
        {
          endogenesData = GetGeneListFromString(value, pawn);
        }
        List<Tuple<Gene, int?, int?>> allGenesData = new List<Tuple<Gene, int?, int?>>();
        foreach (Tuple<Gene, int?, int?> tuple in xenogenesData)
        {
          allGenesData.Add(tuple);
        }
        foreach (Tuple<Gene, int?, int?> tuple in endogenesData)
        {
          allGenesData.Add(tuple);
        }
        foreach (Tuple<Gene, int?, int?> tuple in allGenesData)
        {
          if (tuple.Item2 == null)
          {
            Log.Warning("LoadId was null for a gene");
          }
          if (tuple.Item3 == null)
          {
            continue;
          }
          foreach (Tuple<Gene, int?, int?> otherTuple in allGenesData)
          {
            if (tuple.Item3 != otherTuple.Item2)
            {
              continue;
            }
            if (tuple.Item1 == otherTuple.Item1)
            {
              Log.Warning("Gene tried overriding itself");
              continue;
            }
            tuple.Item1.overriddenByGene = otherTuple.Item1;
            break;
          }
        }

        //////Log.Message("SetGeneTracker, Xenogenes before saving: " + geneTracker.Xenogenes.Join(g => g.def.defName));
        //////Log.Message("SetGeneTracker, Xenogenes to be saved: " + GeneListFromGeneData(xenogenesData).Join(g => g.def.defName));
        Traverse.Create(geneTracker).Field("xenogenes").SetValue(GeneListFromGeneData(xenogenesData));
        //////Log.Message("SetGeneTracker, Xenogenes after saving: " + geneTracker.Xenogenes.Join(g => g.def.defName));

        //////Log.Message("SetGeneTracker, Endogenes before saving: " + geneTracker.Endogenes.Join(g => g.def.defName));
        //////Log.Message("SetGeneTracker, Endogenes to be saved: " + GeneListFromGeneData(endogenesData).Join(g => g.def.defName));
        Traverse.Create(geneTracker).Field("endogenes").SetValue(GeneListFromGeneData(endogenesData));
        //////Log.Message("SetGeneTracker, Endogenes after saving: " + geneTracker.Endogenes.Join(g => g.def.defName));

        if (geneTrackerDict.TryGetValue("xenotype", out value))
        {
          XenotypeDef xenotype = DefDatabase<XenotypeDef>.GetNamedSilentFail(value);
          if (xenotype != null && !value.NullOrEmpty())
          {
            Traverse.Create(geneTracker).Field("xenotype").SetValue(xenotype);
          }
        }

        if (geneTrackerDict.TryGetValue("xenotypeName", out value))
        {
          if (!value.NullOrEmpty() && value != "null")
          {
            geneTracker.xenotypeName = value;
          }
        }

        if (geneTrackerDict.TryGetValue("hybrid", out value))
        {
          if (bool.TryParse(value, out bool hybrid))
          {
            geneTracker.hybrid = hybrid;
          }
        }

        if (geneTrackerDict.TryGetValue("iconDef", out value))
        {
          XenotypeIconDef iconDef = DefDatabase<XenotypeIconDef>.GetNamedSilentFail(value);
          if (iconDef != null && !value.NullOrEmpty())
          {
            geneTracker.iconDef = iconDef;
          }
        }

        Traverse.Create(geneTracker).Field("cachedDamageFactors").SetValue(new Dictionary<DamageDef, float>());
        Traverse.Create(geneTracker).Field("cachedAddictionChanceFactors").SetValue(new Dictionary<ChemicalDef, float>());
        Traverse.Create(geneTracker).Field("cachedGenes").SetValue(null);
        Traverse.Create(geneTracker).Field("cachedGenesAffectAge").SetValue(null);
        Traverse.Create(geneTracker).Field("tmpGeneDefWithTypes").SetValue(new List<GeneDefWithType>());
        Traverse.Create(geneTracker).Field("tmpGenes").SetValue(new List<Gene>());
      }

      ////Log.Message("SetGeneTracker, get melanin gene");
      GeneDef melaninGene = geneTracker.GetMelaninGene();
      if (melaninGene != null && pawn.story != null)
      {
        //////Log.Message("SetGeneTracker, melaninGene = " + melaninGene);
        if (melaninGene.skinColorBase is Color color)
        {
          //////Log.Message("SetGeneTracker, before SkinColorBase = " + pawn.story.SkinColorBase);
          pawn.story.SkinColorBase = color;
          //////Log.Message("SetGeneTracker, after SkinColorBase = " + pawn.story.SkinColorBase);
        }
      }
    }

    public static List<Gene> GeneListFromGeneData(List<Tuple<Gene, int?, int?>> geneDataList)
    {
      List<Gene> geneList = new List<Gene>();
      foreach (Tuple<Gene, int?, int?> tuple in geneDataList)
      {
        geneList.Add(tuple.Item1);
      }
      return geneList;
    }

    // HAIR COLOR GENES
    public static List<GeneDef> GetHairColorGeneDefList(this Pawn pawn)
    {
      List<GeneDef> list = (from g in DefDatabase<GeneDef>.AllDefs
                            where g.endogeneCategory == EndogeneCategory.HairColor
                            select g).ToList();
      list.SortBy(hairGene => pawn.GetHairColorGeneDefLabel(hairGene));
      return list;
    }

    public static string GetHairColorGeneDefLabel(this Pawn pawn, GeneDef hairColorGeneDef = null)
    {
      if (!pawn.HasGeneTracker())
      {
        return "";
      }
      if (hairColorGeneDef == null)
      {
        hairColorGeneDef = pawn.genes.GetHairColorGene();
      }
      string label = hairColorGeneDef.ConstructLabelFromDef().Replace("Hair", "");
      return label.Trim();
    }

    public static void SetHairColorGene(this Pawn pawn, bool next, bool random, bool xenogene = false)
    {
      List<GeneDef> list = pawn.GetHairColorGeneDefList();
      GeneDef newHairColorGeneDef = list.GetNextPrevRandom(pawn.genes.GetHairColorGene(), next, random);
      pawn.SetHairColorGeneDef(newHairColorGeneDef, xenogene);
    }

    public static void SetHairColorGeneDef(this Pawn pawn, GeneDef hairColorGeneDef, bool xenogene = false)
    {
      if (hairColorGeneDef.endogeneCategory != EndogeneCategory.HairColor)
      {
        string error = "SetHairColorGene, tried to set non-hair color gene";
        Log.ErrorOnce(error, error.GetHashCode());
        return;
      }
      pawn.RemoveHairColorGenes();
      pawn.genes.AddGene(hairColorGeneDef, xenogene);
      if (hairColorGeneDef.hairColorOverride is Color hairColor)
      {
        pawn.story.HairColor = hairColor;
      }
    }

    public static void RemoveHairColorGenes(this Pawn pawn)
    {
      foreach (Gene gene in pawn.genes.GenesListForReading)
      {
        if (gene.def.endogeneCategory == EndogeneCategory.HairColor)
        {
          pawn.genes.RemoveGene(gene);
        }
      }
    }

    // SKIN COLOR GENES
    public static List<GeneDef> GetSkinColorGeneDefList(this Pawn pawn)
    {
      List<GeneDef> list = (from g in DefDatabase<GeneDef>.AllDefs
                            where g.endogeneCategory == EndogeneCategory.Melanin
                            select g).ToList();
      list.SortBy(skinColorGeneDef => pawn.GetSkinColorGeneDefLabel(skinColorGeneDef));
      return list;
    }

    public static string GetSkinColorGeneDefLabel(this Pawn pawn, GeneDef skinColorGeneDef = null)
    {
      if (!pawn.HasGeneTracker())
      {
        return "";
      }
      if (skinColorGeneDef == null)
      {
        skinColorGeneDef = pawn.genes.GetMelaninGene();
      }
      string label = skinColorGeneDef.ConstructLabelFromDef().Replace("Skin", "");
      return label.Trim();
    }

    public static void SetSkinColorGene(this Pawn pawn, bool next, bool random, bool xenogene = false)
    {
      List<GeneDef> list = pawn.GetSkinColorGeneDefList();
      GeneDef newSkinColorGeneDef = list.GetNextPrevRandom(pawn.genes.GetMelaninGene(), next, random);
      pawn.SetSkinColorGeneDef(newSkinColorGeneDef, xenogene);
    }



    public static void SetSkinColorGeneDef(this Pawn pawn, GeneDef skinColorGeneDef, bool xenogene = false)
    {
      if (skinColorGeneDef.endogeneCategory != EndogeneCategory.Melanin)
      {
        string error = "SetSkinColorGeneDef, tried to set gene without EndogeneCategory.Melanin";
        Log.ErrorOnce(error, error.GetHashCode());
        return;
      }
      pawn.RemoveSkinColorGenes();
      pawn.genes.AddGene(skinColorGeneDef, xenogene);

      if (skinColorGeneDef.skinColorBase is Color skinColor)
      {
        pawn.story.SkinColorBase = skinColor;
      }
      pawn.story.skinColorOverride = skinColorGeneDef.skinColorOverride;
    }

    public static void RemoveSkinColorGenes(this Pawn pawn)
    {
      foreach (Gene gene in pawn.genes.GenesListForReading)
      {
        if (gene.def.endogeneCategory == EndogeneCategory.Melanin)
        {
          pawn.genes.RemoveGene(gene);
        }
      }
    }


  }
}

